/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File vgaevents.c : Event handling code - Linux VGA version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <vga.h>
#include <vgamouse.h>
#include <vgakeyboard.h>

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "vgasamP.h"
#include "resource.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "vgaspectrum.h"

int * key_table;

/* Maps vgalib scancodes to a Sam keyboard row and bit to set/reset */
/* E.g. ktbl[SCANCODE_A] means set/reset bit 0 (0xfe) of keyboard row 1 */
struct ktbl {
	int port;	/* which keyboard row is the key on? */
	int bit;	/* which bit to mask */
} keytable[51] = {
	{9, 0}, {9, 0}, 
	/* 1 2 3 4 5 6 7 8 9 0 */
        {3, 0xfe}, {3, 0xfd}, {3, 0xfb}, {3, 0xf7}, {3, 0xef}, 
        {4, 0xef}, {4, 0xf7}, {4, 0xfb}, {4, 0xfd}, {4, 0xfe},
                	
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* Q W E R T Y U I O P */
	{2, 0xfe}, {2, 0xfd}, {2, 0xfb}, {2, 0xf7}, {2, 0xef},
	{5, 0xef}, {5, 0xf7}, {5, 0xfb}, {5, 0xfd}, {5, 0xfe},
	
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* A S D F G H J K L */
	{1, 0xfe}, {1, 0xfd}, {1, 0xfb}, {1, 0xf7}, {1, 0xef},
	{6, 0xef}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd},
	
	{9, 0}, {9, 0}, {9, 0}, {9,0}, {9,0},

	/* Z X C V B N M */
	           {0, 0xfd}, {0, 0xfb}, {0, 0xf7}, {0, 0xef},
	{7, 0xef}, {7, 0xf7}, {7, 0xfb}
	           	
};

static struct k_map {
	char *res_name;
	int  Scancode;
} map_tab[NUM_KEYMAPPINGS] = {
	{ "JoystUp",    SCANCODE_9 }, 
	{ "JoystDown",  SCANCODE_8 },
	{ "JoystLeft",  SCANCODE_6 },
	{ "JoystRight", SCANCODE_7 },
	{ "JoystFire",  SCANCODE_0 },
#ifdef SLOWDOWN
	{ "keySlower",   SCANCODE_KEYPADMINUS },
	{ "keyFaster",   SCANCODE_KEYPADPLUS },
#endif
};

uns8 keyports[10];	/* 8 rows of keys (+ 1 row for unscanned keys) */
uns8 inPorts[256], outPorts[256];

/* All to do with screen painting */

uns8 *vga_data_ptr;			/* pointer to actual frame buffer data */		

extern int ClutChangeInFrame;		/* signals a clut change has been made in 
					   the present frame */
extern int BorderChangeInFrame;		/* signals a border change has been made in
					   the present frame */
extern int VmprChangeInFrame;		/* signals a vmpr change has been made in
					   the present frame */

int SpectrumKeyboard;			/* Switch for spectrum keyboard
					   mode - see vgaspectrum.c */
static int Vga2x2ModeAvailable;		/* Flags for the 2x2 display mode */
int Vga2x2LinearAddressing;
static int ScaleFactor;			/* Mode for dynamic scale switching */
						  
static void SamKeyboardHandler(int Scancode, int EventType);
static void SamMouseHandler(int Button, int dx, int dy);


extern int flash_phase;
extern uns8 LineNeedsUpdating[];

extern void (*line_update)(int);
extern void (*line_update_mode_1)(int);
extern void (*line_update_mode_2)(int);
extern void (*line_update_mode_3)(int);
extern void (*line_update_mode_4)(int);

void call_vmpr(uns8 val);

static int SAMButtons  = 0xff;  	/* Mouse button status */
int SAMPointerMode = 0;			/* Pointer Attatched and x magnification */

/* from processor.c */
extern int LineNo;
extern int disass;
extern int InterruptType;
 
/* The SAM ports */

uns8 vmpr = 0x81;
uns8 hmpr = 0x01;
uns8 lmpr = 0x1f;
uns8 midi = 0x00;	
uns8 border = 0x00;
uns8 clut[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 
                 0x00, 0x19, 0x2a, 0x3b, 0x4c, 0x5d, 0x6e, 0x7f};
uns8 keyboard = 0x5f;
uns8 status_reg = 0xff;
uns8 line_int = 0xff;

uns8 hpen = 0x00;
uns8 lpen = 0x00;
        	
uns8 clock_reg = 0xff;

uns8 lepr = 0x00;
uns8 hepr = 0x00;

uns8 attr_reg = 0xff;

/*
 *  Private functions start here
 */

/* Perform a clean close-down. */

static void
closedown(void)
{
	mouse_close();
	keyboard_close();
	vga_setmode(TEXT);
}


#ifdef SLOWDOWN
/* Speed up or slow down the emulator */
static void
adjust_speed(int delta)
{
	int newspd = GetCfg(slowdown) + delta;

	if (newspd < 0) newspd = GetCfg(slowdown);

	SetCfg(slowdown, newspd);
}
#endif


static int ShiftPressed = False;		/* Flag for shift key */
static int LastKeyScancode = 0xff;		/* Use to stop auto repeat */

/* Process a KeyPress event */
static void
process_keypress(int ks)
{
	char *str;
	int i;	

	if (ks == LastKeyScancode && 
	    (ks != SCANCODE_LEFTSHIFT || ks != SCANCODE_RIGHTSHIFT))
		return;
		 
	LastKeyScancode = ks;
	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] |= 0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] |= 0x01;
#ifdef SLOWDOWN
	else if (ks == key_table[KM_KEY_SLOWER])
		adjust_speed(1);
	else if (ks == key_table[KM_KEY_FASTER])
		adjust_speed(-1);
#endif

	switch (ks) {
		/* Emulator control keys */
		case SCANCODE_F1: 
			InterruptType = Z80_reset;
			for (i = 0; i < 9; i++)
				keyports[i] = 0xff;
			break;
		case SCANCODE_F2:	/* Switch Keyboard mode */
			SpectrumKeyboard = !SpectrumKeyboard;
			break;
		case SCANCODE_F3:	/* Reset Mouse / pointer */ 
			if (!SAMPointerMode) {
				SAMButtons  = 0xff;
			}
			SAMPointerMode = (SAMPointerMode + 1) % 3;
			break;

		case SCANCODE_F4:
			break;
		case SCANCODE_F5:	/* Rescale windows */
			if (Vga2x2ModeAvailable) {
				ScaleFactor = (ScaleFactor == 1) ? 2 : 1;
				ChangeScaleMode(ScaleFactor);

				/* Call vmpr to reset the screen write pointers and
				   fetch the screen from memory */
				call_vmpr(vmpr);
			}	
			break;

		case SCANCODE_F6:     /* Insert new floppy disk in drive 1 */
			str = dialog_event_loop("Disk Image for drive 1");
			if (strlen(str))
				insert_disk(str, 1);
			/* Reset all the graphics properly */
			ChangeScaleMode(ScaleFactor);
			break;
		case SCANCODE_F7:     /* Insert new floppy disk in drive 2 */
			str = dialog_event_loop("Disk Image for drive 2");
			if (strlen(str))
				insert_disk(str, 2);
			/* Reset all the graphics properly */
			ChangeScaleMode(ScaleFactor);
			break;
		case SCANCODE_F8:	/* Generate an NMI */
			InterruptType = Z80_nmi;
			break;
			
		/* Crude Debug toggle */
		case SCANCODE_F9:
#if 0
			str = dialog_event_loop("Toggle debug");
			disass = !disass;
#endif
			break;
		case SCANCODE_F10: case SCANCODE_F12:
			Quit(0);
			break;

		/* For the shift key we need to set a flag */
		case SCANCODE_LEFTSHIFT: case SCANCODE_RIGHTSHIFT:
			keyports[0] &= 0xfe; 
			ShiftPressed = True;
			break;


		/* Ordinary SAM keys */

		case SCANCODE_SPACE:
			keyports[7] &= 0xfe; break;
		case SCANCODE_TAB:
			keyports[3] &= 0xbf; break;
		case SCANCODE_ENTER:
			keyports[6] &= 0xfe; break;
		case SCANCODE_LEFTCONTROL: case SCANCODE_RIGHTCONTROL:
		/* <Shift> <Control> is mapped to Caps Lock */
			if (keyports[0] & 0x01)
			     keyports[8] &= 0xfe;
			else
			     keyports[3] &= 0x7f;
			break;
		case SCANCODE_LEFTALT:
/*		case SCANCODE_LEFTMETA: ????? */
			keyports[7] &= 0xfd; break;

		case SCANCODE_RIGHTALT: 
/*		case SCANCODE_RIGHTMETA: case SCANCODE_Mode_switch: ????? */
			keyports[6] &= 0x7f; break;
		case SCANCODE_BACKSPACE: 
/*		case SCANCODE_DELETE: ????? */
			keyports[4] &= 0x7f; break;
		case SCANCODE_ESCAPE:
			keyports[3] &= 0xdf; break;
		case SCANCODE_CURSORBLOCKUP:
			keyports[8] &= 0xfd; break;
		case SCANCODE_CURSORBLOCKDOWN:
			keyports[8] &= 0xfb; break;
		case SCANCODE_CURSORBLOCKLEFT:
			keyports[8] &= 0xf7; break;
		case SCANCODE_CURSORBLOCKRIGHT:
			keyports[8] &= 0xef; break;
		case SCANCODE_MINUS:
			if (!ShiftPressed)   
				keyports[4] &= 0xdf;
			else /* SCANCODE_UNDERSCORE */
				keyports[5] &= 0xdf;
			break;
		case SCANCODE_EQUAL:
			if (!ShiftPressed)   
				keyports[5] &= 0xdf;
			else { /* SCANCODE_PLUS */
				keyports[0] |= 0x01; 
				keyports[4] &= 0xbf;
			}
			break;
		case SCANCODE_SEMICOLON:
			if (!ShiftPressed)   
				keyports[6] &= 0xdf;
			else { /* SCANCODE_COLON */
				keyports[0] |= 0x01; 
				keyports[6] &= 0xbf;
			}
			break;
		case SCANCODE_APOSTROPHE:
			if (!ShiftPressed) { 
				keyports[0] &= 0xfe; 
				keyports[4] &= 0xf7;
			}
			else { /* SCANCODE_AT */
				keyports[3] &= 0xfd;
			}
			break;
		case SCANCODE_COMMA:
			if (!ShiftPressed) 
				keyports[7] &= 0xdf;
			else { /* SCANCODE_LESS */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[2] &= 0xfe; 
			}
			break;	
		case SCANCODE_PERIOD:
			if (!ShiftPressed) 
				keyports[7] &= 0xbf;
			else { /* SCANCODE_GREATER */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[2] &= 0xfd;
			}
			break;
		case SCANCODE_SLASH:
			if (!ShiftPressed) {
				keyports[0] &= 0xfe; 
				keyports[4] &= 0xdf;
			}
			else { /* SCANCODE_QUESTION */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[0] &= 0xfb;
			}
			break;
		case SCANCODE_UK_BACKSLASH:
			if (!ShiftPressed) {
			        keyports[0] &= 0xfe; keyports[7] &= 0x7f;
			}
			else {	/* SCANCODE_BAR */	
			        keyports[0] |= 0x01; keyports[7] &= 0xfd;
			        keyports[4] &= 0xfd;
			}
			break;
                case SCANCODE_7:
			if (!ShiftPressed) 
				keyports[4] &= 0xf7;
			else /* SCANCODE_AMPERSAND */                
	                        keyports[4] &=0xef; 
	                break;
                case SCANCODE_BRACKET_LEFT:
			if (!ShiftPressed) {
	                	keyports[7] &= 0xfd; keyports[2] &= 0xf7;
			}
			else { /* SCANCODE_BRACERIGHT */
	                        keyports[0] |= 0x01;
	                        keyports[7] &= 0xfd; keyports[1] &= 0xf7;
			}
			break;
                case SCANCODE_BRACKET_RIGHT:
			if (!ShiftPressed) {
				keyports[7] &= 0xfd; keyports[2] &= 0xef;
			}
			else { /* SCANCODE_BRACERIGHT */
	                        keyports[0] |= 0x01;
				keyports[7] &= 0xfd; keyports[1] &= 0xef;
			}
			break;
		case SCANCODE_2:
			if (!ShiftPressed) 
				keyports[3] &= 0xfd;
			else { /* SCANCODE_QUOTEDBL */
				keyports[0] |= 0x01; 
				keyports[5] &= 0xbf;
			}			
			break;

		case SCANCODE_UK_HASH:   /* NB Not in vgakeyboard.h */
			if (!ShiftPressed) {
	                        keyports[0] |= 0x01;
				keyports[0] &= 0xfe; keyports[3] &=0xfb;
			}
			else   /* SCANCODE_ASCIITILDE */ 
				keyports[4] &= 0xfe;
			break;

		case SCANCODE_6:
			if (!ShiftPressed) {
				keyports[4] &= 0xef;
			}
			else { /* SCANCODE_ASCIICIRCUM */
			        keyports[0] |= 0x01; 
			        keyports[7] &= 0xfd;
			        keyports[6] &= 0xef;
			}
		        break;

		case SCANCODE_3:
			if (!ShiftPressed) {
				keyports[3] &= 0xfb;
			}
			else { /* SCANCODE_STERLING */
	                        keyports[0] |= 0x01; 
				keyports[7] &= 0xfd; keyports[6] &= 0xfd;
			}
			break;

		case SCANCODE_8:
			if (!ShiftPressed) {
				keyports[4] &= 0xfb;
			}
			else { /* SCANCODE_ASTERISK */
				keyports[4] &= 0xbf; 
			}
			break;

		case SCANCODE_9:
			if (!ShiftPressed) {
				keyports[4] &= 0xfd;
			}
			else { /* SCANCODE_LEFTPAR */
				keyports[4] &= 0xfb;
			}
			break;

		case SCANCODE_0:
			if (!ShiftPressed) {
				keyports[4] &= 0xfe;
			}
			else { /* SCANCODE_RIGHTPAR */
				keyports[4] &= 0xfd;
			}
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case SCANCODE_KEYPAD0:
			keyports[5] &= 0x7f;
			break;
		case SCANCODE_KEYPAD1:
			keyports[0] &= 0xdf;
			break;
		case SCANCODE_KEYPAD2:
			keyports[0] &= 0xbf;
			break;
		case SCANCODE_KEYPAD3:
			keyports[0] &= 0x7f;
			break;
		case SCANCODE_KEYPAD4:
			keyports[1] &= 0xdf;
			break;
		case SCANCODE_KEYPAD5:
			keyports[1] &= 0xbf;
			break;
		case SCANCODE_KEYPAD6:
			keyports[1] &= 0x7f;
			break;
		case SCANCODE_KEYPAD7:
			keyports[2] &= 0xdf;
			break;
		case SCANCODE_KEYPAD8:
			keyports[2] &= 0xbf;
			break;
		case SCANCODE_KEYPAD9:
			keyports[2] &= 0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case SCANCODE_KEYPADENTER:
                        keyports[7] &= 0x7f;
                	break;
                case SCANCODE_KEYPADPERIOD:
			keyports[0] &= 0xfe; keyports[5] &= 0xbf;
                	break;

		default:
			if (ks < 52) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
static void
process_keyrelease(int ks)
{
	char *str;
	int i;	

	if (ks == LastKeyScancode)
		LastKeyScancode = 0xff;

	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] &= ~0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] &= ~0x01;
#ifdef SLOWDOWN
	else if (ks == key_table[KM_KEY_SLOWER])
		adjust_speed(1);
	else if (ks == key_table[KM_KEY_FASTER])
		adjust_speed(-1);
#endif

	switch (ks) {

		/* For the shift key we need to reset a flag */
		case SCANCODE_LEFTSHIFT: case SCANCODE_RIGHTSHIFT:
			keyports[0] |= ~0xfe; 
			ShiftPressed = False;
			break;

		/* Ordinary SAM keys */
		case SCANCODE_SPACE:
			keyports[7] |= ~0xfe; break;
		case SCANCODE_TAB:
			keyports[3] |= ~0xbf; break;
		case SCANCODE_ENTER:
			keyports[6] |= ~0xfe; break;
		case SCANCODE_LEFTCONTROL: case SCANCODE_RIGHTCONTROL:
		/* <Shift> <Control> is mapped to Caps Lock */
			keyports[8] |= ~0xfe;
	     		keyports[3] |= ~0x7f;
			break;
		case SCANCODE_LEFTALT:
/*		case SCANCODE_LEFTMETA: ????? */
			keyports[7] |= ~0xfd; break;

		case SCANCODE_RIGHTALT: 
/*		case SCANCODE_RIGHTMETA: case SCANCODE_Mode_switch: ????? */
			keyports[6] |= ~0x7f; break;
		case SCANCODE_BACKSPACE: 
/*		case SCANCODE_DELETE: ????? */
			keyports[4] |= ~0x7f; break;
		case SCANCODE_ESCAPE:
			keyports[3] |= ~0xdf; break;
		case SCANCODE_CURSORBLOCKUP:
			keyports[8] |= ~0xfd; break;
		case SCANCODE_CURSORBLOCKDOWN:
			keyports[8] |= ~0xfb; break;
		case SCANCODE_CURSORBLOCKLEFT:
			keyports[8] |= ~0xf7; break;
		case SCANCODE_CURSORBLOCKRIGHT:
			keyports[8] |= ~0xef; break;
		case SCANCODE_MINUS:
			if (!ShiftPressed)   
				keyports[4] |= ~0xdf;
			else /* SCANCODE_UNDERSCORE */
				keyports[5] |= ~0xdf;
			break;
		case SCANCODE_EQUAL:
			if (!ShiftPressed)   
				keyports[5] |= ~0xdf;
			else { /* SCANCODE_PLUS */
				keyports[0] &= ~0x01; 
				keyports[4] |= ~0xbf;
			}
			break;
		case SCANCODE_SEMICOLON:
			if (!ShiftPressed)   
				keyports[6] |= ~0xdf;
			else { /* SCANCODE_COLON */
/*				keyports[0] &= ~0x01; */
				keyports[6] |= ~0xbf;
			}
			break;
		case SCANCODE_APOSTROPHE:
			if (!ShiftPressed) { 
				keyports[0] |= ~0xfe; 
				keyports[4] |= ~0xf7;
			}
			else { /* SCANCODE_AT */
				keyports[3] |= ~0xfd;
			}
			break;
		case SCANCODE_COMMA:
			if (!ShiftPressed) 
				keyports[7] |= ~0xdf;
			else { /* SCANCODE_LESS */
/*				keyports[0] &= ~0x01; */
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfe; 
			}
			break;	
		case SCANCODE_PERIOD:
			if (!ShiftPressed) 
				keyports[7] |= ~0xbf;
			else { /* SCANCODE_GREATER */
/*				keyports[0] &= ~0x01; */
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfd;
			}
			break;
		case SCANCODE_SLASH:
			if (!ShiftPressed) {
				keyports[0] |= ~0xfe; 
				keyports[4] |= ~0xdf;
			}
			else { /* SCANCODE_QUESTION */
/*				keyports[0] &= ~0x01; */
				keyports[7] |= ~0xfd;
				keyports[0] |= ~0xfb;
			}
			break;
		case SCANCODE_UK_BACKSLASH:
			if (!ShiftPressed) {
			        keyports[0] |= ~0xfe; keyports[7] |= ~0x7f;
			}
			else {	/* SCANCODE_BAR */	
/*			        keyports[0] &= ~0x01; */
			        keyports[7] |= ~0xfd;
			        keyports[4] |= ~0xfd;
			}
			break;
                case SCANCODE_7:
			if (!ShiftPressed) 
				keyports[4] |= ~0xf7;
			else /* SCANCODE_AMPERSAND */                
	                        keyports[4] |= ~0xef; 
	                break;
                case SCANCODE_BRACKET_LEFT:
			if (!ShiftPressed) {
	                	keyports[7] |= ~0xfd; keyports[2] |= ~0xf7;
			}
			else { /* SCANCODE_BRACELEFT */
	                        keyports[0] &= ~0x01; 
	                        keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
			}
			break;
                case SCANCODE_BRACKET_RIGHT:
			if (!ShiftPressed) {
				keyports[7] |= ~0xfd; keyports[2] |= ~0xef;
			}
			else { /* SCANCODE_BRACERIGHT */
	                        keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd; keyports[1] |= ~0xef;
		}
			break;
		case SCANCODE_2:
			if (!ShiftPressed) 
				keyports[3] |= ~0xfd;
			else { /* SCANCODE_QUOTEDBL */
				keyports[0] &= ~0x01; 
				keyports[5] |= ~0xbf;
			}			
			break;

		case SCANCODE_UK_HASH:   /* NB Not in vgakeyboard.h */
			if (!ShiftPressed) {
/*	                        keyports[0] &= ~0x01; */
				keyports[0] |= ~0xfe; keyports[3] |= ~0xfb;
			}
			else   /* SCANCODE_ASCIITILDE */ 
				keyports[4] |= ~0xfe;
			break;

		case SCANCODE_6:
			if (!ShiftPressed) {
				keyports[4] |= ~0xef;
			}
			else { /* SCANCODE_ASCIICIRCUM */
			        keyports[0] &= ~0x01; 
			        keyports[7] |= ~0xfd;
			        keyports[6] |= ~0xef;
			}
		        break;

		case SCANCODE_3:
			if (!ShiftPressed) {
				keyports[3] |= ~0xfb;
			}
			else { /* SCANCODE_STERLING */
	                        keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd; 
				keyports[6] |= ~0xfd;
			}
			break;

		case SCANCODE_8:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfb;
			}
			else { /* SCANCODE_ASTERISK */
				keyports[4] |= ~0xbf; 
			}
			break;

		case SCANCODE_9:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfd;
			}
			else { /* SCANCODE_LEFTPAR */
				keyports[4] |= ~0xfb;
			}
			break;

		case SCANCODE_0:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfe;
			}
			else { /* SCANCODE_RIGHTPAR */
				keyports[4] |= ~0xfd;
			}
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case SCANCODE_KEYPAD0:
			keyports[5] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD1:
			keyports[0] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD2:
			keyports[0] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD3:
			keyports[0] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD4:
			keyports[1] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD5:
			keyports[1] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD6:
			keyports[1] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD7:
			keyports[2] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD8:
			keyports[2] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD9:
			keyports[2] |= ~0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case SCANCODE_KEYPADENTER:
                        keyports[7] |= ~0x7f;
                	break;
                case SCANCODE_KEYPADPERIOD:
			keyports[0] |= ~0xfe; keyports[5] |= ~0xbf;
                	break;

		default:
			if (ks < 52) {
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			}
			break;
	}
}

/*
 * Display initialisation routines
 */

static int
mode_init(void)
{
	vga_modeinfo *mode_info;

	if (Vga2x2ModeAvailable)
		mode_info = vga_getmodeinfo(VGA2X2MODE);
	else
		mode_info = vga_getmodeinfo(VGA1X1MODE);
		
	return 0;
}


/*
 * Initialise the 128 colours required by the SAM
 * Return -1 if colour allocation failed.
 */
int
get_colours(void)
{
	int i;
        int red_index, green_index, blue_index;

	int RGBvals[8] = {0x00, 0x09, 0x12, 0x1b,  
	                  0x24, 0x2d, 0x36, 0x3f};


	for (i = 0; i < NCOLOURS; i++) {

		blue_index  = (i & 0x01) << 1| ((i & 0x10) >> 2) | ((i & 0x08) >> 3);                       
		red_index   = (i & 0x02)     | ((i & 0x20) >> 3) | ((i & 0x08) >> 3);                       
		green_index = (i & 0x04) >> 1| ((i & 0x40) >> 4) | ((i & 0x08) >> 3);                       

		if (vga_setpalette(i, RGBvals[red_index],
				      RGBvals[green_index],
				      RGBvals[blue_index])) {
			simcoupe_mesg(SIMCOUPE_WARN, "get_colours() failed on colour %d", i);
			return(-1);
		}
		pixels[i] = (long)i;
	}
	return 0;
}


/*
 *  Public functions start here
 */


void init_key(void)
{
	if (keyboard_init()) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Can't initialise keyboard");
	} 	

	/* Set the keyboard handler - so we can process key press AND release */
	
	keyboard_seteventhandler(SamKeyboardHandler); 


}

void
init_display(void)
{
	char *MouseDevice	= MOUSE_DEVICE;
	int   MouseType		= vga_getmousetype();
	int   MouseSampleRate	= 0;

	int i;

	/* Initialise the keysym -> SAM action translations */

	key_table = (int *)malloc(sizeof(int) * NUM_KEYMAPPINGS);
	for (i = 0; i < NUM_KEYMAPPINGS; i++) {
		key_table[i] = map_tab[i].Scancode;
	}


	/* initialise svgalib */
	
	vga_init();
	
	/* Test to see what vga modes are available */
	
	if (!vga_hasmode(VGA1X1MODE))
                simcoupe_mesg(SIMCOUPE_FATAL, "Failed to find 320x200x256 SVGA mode");
	if (vga_hasmode(VGA2X2MODE)) { 
		simcoupe_mesg(SIMCOUPE_INFO, "Found 640x480x256 SVGA mode, testing"
					 " for linear addressing");
		Vga2x2ModeAvailable = True;
		vga_setmode(VGA2X2MODE);
		if (vga_setlinearaddressing() != -1)
			Vga2x2LinearAddressing = True;
		else
			Vga2x2LinearAddressing = False;
		vga_setmode(TEXT);

		if (Vga2x2LinearAddressing)
			simcoupe_mesg(SIMCOUPE_INFO,"Linear addressing available");
		else
			simcoupe_mesg(SIMCOUPE_INFO,"Linear addressing not available - using paging");
				
	}
	else {
		simcoupe_mesg(SIMCOUPE_WARN, "Failed to find 640x480x256 SVGA mode"
					    " : only 1x1 scaling will be available");	
		Vga2x2ModeAvailable = False;
		Vga2x2LinearAddressing = False;
	}

	
	flash_phase = !GetCfg(flashing);


	/* Initialise the colourmap - NB we have to redo this everytime
	   the vga mode is changed! - here we are just testing the colours
	   are all available */
	if (get_colours() == -1) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Colourmap initialisation failed!");
	}


	/* Initialise variables used to keep track of line interrupt 
	   clut changes */

	ClutChangeInFrame   = 0;
	BorderChangeInFrame = 0;
	VmprChangeInFrame   = 0;	

	/* Set the dynamic scaling to 1x1 initially */

	ScaleFactor = 1;
	ChangeScaleMode(ScaleFactor);
	line_update = line_update_mode_1;

	/* Image handling initialisation */
	if (mode_init() != 0) {
		simcoupe_mesg(SIMCOUPE_FATAL, "vga mode initialisation failure!");
	}

	/* Keyboard initialisation */
	
	init_key();

	/* No keys initially being pressed */
	for (i = 0; i < 9; i++)
		keyports[i] = 0xff;
	/* Keyboard starts off in SAM mode by default */
	SpectrumKeyboard = 0;

/* Audio initialisation would go here, but none for now */ 

#ifdef JOY
	/* Linux analogue joystick initialisation */
	if (GetCfg(use_js) && joystick_init() != 0) { 
		simcoupe_mesg(SIMCOUPE_WARN, "joystick initialisation failed");
		SetCfg(use_js, 0);
	}
#endif

	/* Mouse initialisation */

	if (mouse_init(MouseDevice, MouseType, MouseSampleRate)) {
		simcoupe_mesg(SIMCOUPE_WARN,"Can't initialise mouse");
	} 	
	mouse_seteventhandler(SamMouseHandler);
	sam_mouse_init();

	OnQuit(closedown);
}


/* Check & process any incoming events */
void 
check_events(void)
{
	/* Actually all we do is to call the svgalib mouse and
	   keyboard update routines - which in turn call the
	   appropriate update routines */
	   
	if (SAMPointerMode)
		sam_mouse_xy_write(0, 0);

	mouse_update();
	keyboard_update();
	
}

/* Keyboard handler - control is passed to here if an event is
   pending when keyboard_update() is called from check_events() */

static void
SamKeyboardHandler(int Scancode, int EventType)
{
 
	switch (EventType) {
	
		case KEY_EVENTPRESS:
			if (!SpectrumKeyboard)
				process_keypress(Scancode);
			else
				process_spectrum_keypress(Scancode);
			break;
		
		case KEY_EVENTRELEASE:
			if (!SpectrumKeyboard)
				process_keyrelease(Scancode);
			else
				process_spectrum_keyrelease(Scancode);
			break;
		default:
			simcoupe_mesg(SIMCOUPE_ERR,"Unknown Key event type");	

	} /* end switch */

}

/* Mouse handler - control is passed to here if an event is
   pending when mouse_update() is called from check_events() */

static void
SamMouseHandler(int Button, int dx, int dy)
{
	static int OldButton = 0xff;

	/* process a pointer move event */
	if (dx || dy) { 
		dx = (dx*SAMPointerMode);
		dy = - dy;
		sam_mouse_xy_write(dx, dy);
	}

	if (OldButton != Button) {

		OldButton = Button;

		if (Button & MOUSE_LEFTBUTTON)
			SAMButtons &= ~SAM_BUTTON_1;
		else
			SAMButtons |=  SAM_BUTTON_1;
	
		if (Button & MOUSE_MIDDLEBUTTON)
			SAMButtons &= ~SAM_BUTTON_2;
		else
			SAMButtons |=  SAM_BUTTON_2;
	
		if (Button & MOUSE_RIGHTBUTTON)
			SAMButtons &= ~SAM_BUTTON_3;
		else
			SAMButtons |=  SAM_BUTTON_3;

		sam_mouse_button_write(SAMButtons);
	}	
}	
